# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from decimal import Decimal
from typing import List, Union

import aiohttp

from obm.connectors import base

__all__ = [
    "BitcoinCoreConnector",
]


class BitcoinCoreConnector(base.Connector):
    node = "bitcoin-core"
    currency = "bitcoin"

    # TODO: Migrate to __slots__
    METHODS = {
        "rpc_list_transactions": "listtransactions",
        "rpc_estimate_smart_fee": "estimatesmartfee",
        "rpc_send_to_address": "sendtoaddress",
        "rpc_get_new_address": "getnewaddress",
        "rpc_get_block_count": "getblockcount",
        "rpc_get_block": "getblock",
        "rpc_get_transaction": "gettransaction",
        "rpc_get_raw_transaction": "getrawtransaction",
    }

    def __init__(
        self,
        rpc_host: str = "localhost",
        rpc_port: int = 18332,
        rpc_username: str = None,
        rpc_password: str = None,
        loop=None,
        session: aiohttp.ClientSession = None,
        timeout: int = None,
    ):
        if rpc_username is not None and rpc_password is not None:
            self.auth = aiohttp.BasicAuth(rpc_username, rpc_password)
        # self.serializer = serializers.BitcoinCoreTransaction()
        self.headers = {
            "content-type": "application/json",
            "cache-control": "no-cache",
        }
        super().__init__(rpc_host, rpc_port, loop, session, timeout)

    async def wrapper(self, *args, method: str = None) -> Union[dict, list]:
        assert method is not None
        response = await self.call(payload={"method": method, "params": args,})
        return await self.validate(response)

    # BitcoinCore specific interface

    @staticmethod
    def format_transaction(tx, latest_block_number):
        category = tx.get("category")
        if category is None:
            # Tx is send_transaction output.
            details = {detail["category"]: detail for detail in tx["details"]}
            fee = details["send"]["fee"]
            amount = details["send"]["amount"]
            to_address = details["receive"]["address"]
            from_address = details["send"]["address"]
            category = "send" if from_address != to_address else "oneself"
        else:
            # Tx is list_transactions output.
            if category == "oneself":
                from_address = to_address = tx["address"]
                # Recover original category to prevent lie in info key.
                tx["category"] = "send"
            else:
                to_address = tx["address"]
                from_address = None
            amount = tx["amount"]
            fee = tx.get("fee", 0)

        if tx["confirmations"] == -1:
            # This mean that tx out of the main chain.
            # Reference: https://bitcoin.org/en/developer-reference#listtransactions
            block_number = -1
        elif tx["confirmations"] == 0:
            # Transaction  still in mempool.
            block_number = None
        else:
            number_from_end = tx["confirmations"] - 1
            block_number = latest_block_number - number_from_end

        return {
            "txid": tx["txid"],
            "from_address": from_address,
            "to_address": to_address,
            "amount": Decimal(str(abs(amount))),
            "fee": Decimal(str(abs(fee))),
            "block_number": block_number,
            "category": category,
            "timestamp": tx["time"],
            "info": tx,
        }

    # Unified interface

    @property
    async def latest_block_number(self) -> int:
        return await self.rpc_get_block_count()

    async def create_address(  # pylint: disable=unused-argument
        self, password: str = ""
    ) -> str:
        # TODO: Add args
        return await self.rpc_get_new_address()

    async def estimate_fee(  # pylint: disable=unused-argument
        self, transaction: dict = None, conf_target: int = 1,
    ) -> Decimal:
        fee_estimate = await self.rpc_estimate_smart_fee(conf_target)
        return Decimal(str(fee_estimate["feerate"]))

    async def send_transaction(  # pylint: disable=unused-argument
        self,
        amount: Decimal,
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
    ) -> dict:
        # TODO: Validate
        latest_block_number = await self.latest_block_number
        txid = await self.rpc_send_to_address(to_address, amount)
        tx = await self.rpc_get_transaction(txid)
        return self.format_transaction(tx, latest_block_number)

    async def list_transactions(self, count: int = 10, **kwargs) -> List[dict]:
        """Lists most recent transactions.

        Args:
            count: The number of transactions to return. Defaults to 10.

        Returns:
            Recent transactions list.
        """

        def combine_duplicates(txs):
            """Combines doubles that is transaction self-sending outcome."""
            txs_by_txid = defaultdict(list)
            for tx in txs:
                txs_by_txid[tx["txid"]].append(tx)
            duplicate_txids = [
                txid
                for txid, tx_or_txs in txs_by_txid.items()
                if len(tx_or_txs) == 2
            ]
            txs = [tx for tx in txs if tx["txid"] not in duplicate_txids]
            for duplicate_txid in duplicate_txids:
                duplicate_txs = {
                    tx["category"]: tx for tx in txs_by_txid[duplicate_txid]
                }
                assert "send" in duplicate_txs and "receive" in duplicate_txs
                duplicate_txs["send"]["category"] = "oneself"
                txs.append(duplicate_txs["send"])
            return txs

        # Double increase the transactions number for feching to prevent cases
        # when transaction was sent on in-wallet address and is present in
        # bitcoin core list tansaction twice with different categories.
        inner_count = count * 2
        latest_block_number = await self.latest_block_number
        txs = await self.rpc_list_transactions(
            kwargs.get("label", "*"),
            inner_count,
            kwargs.get("skip", 0),
            kwargs.get("include_watchonly", False),
        )
        sorted_txs = sorted(
            [
                self.format_transaction(tx, latest_block_number)
                for tx in combine_duplicates(txs)
            ],
            key=lambda x: x["timestamp"],
            reverse=True,
        )
        return sorted_txs[:count]



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from decimal import Decimal
from typing import List, Union

from obm import connectors


class ConnectorMixin:

    __connector = None

    @property
    def connector(self):
        if self.__connector:
            return self.__connector
        self.__connector = connectors.MAPPING[self.name](
            self.rpc_host,
            self.rpc_port,
            self.rpc_username,
            self.rpc_password,
            self.loop,
            self.session,
        )
        return self.__connector

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def open(self):
        await self.connector.close()

    async def close(self):
        await self.connector.close()

    async def create_address(self, password: str = "") -> str:
        return await self.connector.create_address(password)

    async def estimate_fee(self, transaction: dict = None) -> Decimal:
        return await self.connector.estimate_fee(transaction)

    async def list_transactions(self, count: int = 10) -> List[dict]:
        return await self.connector.list_transactions(count)

    async def send_transaction(
        self,
        amount: Decimal,
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
    ) -> dict:
        return await self.connector.send_transaction(
            amount, to_address, from_address, fee, password
        )



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import aiohttp

from obm import connectors, mixins

__all__ = [
    "Currency",
    "Node",
]


class Currency:
    def __init__(self, name: str, symbol: str = None):
        self.name = name
        self.symbol = symbol

    @classmethod
    def create_for(cls, connector_name: str):
        return cls(name=connectors.MAPPING[connector_name].currency)


class Node(mixins.ConnectorMixin):
    def __init__(
        self,
        name: str,
        rpc_port: int,
        currency: Currency = None,
        rpc_host: str = "127.0.0.1",
        rpc_username: str = None,
        rpc_password: str = None,
        loop=None,
        session: aiohttp.ClientSession = None,
    ):
        self.name = name
        self.currency = currency or Currency.create_for(name)
        self.rpc_port = rpc_port
        self.rpc_host = rpc_host
        self.rpc_username = rpc_username
        self.rpc_password = rpc_password
        self.loop = loop
        self.session = session
        super().__init__()

    # @staticmethod
    # def validate_name(name: str) -> str:
    #     if not isinstance(name, str):
    #         raise TypeError(
    #             f"Name argument must be a string, not '{type(name)}'"
    #         )
    #     supported_nodes = list(connectors.MAPPING)
    #     if name not in supported_nodes:
    #         raise ValueError(
    #             f"Unsupported node. Available only: {supported_nodes}"
    #         )
    #     return name



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from decimal import Decimal
from typing import List, Union

import aiohttp

from obm import utils
from obm.connectors import base

__all__ = [
    "GethConnector",
]


class GethConnector(base.Connector):
    node = "geth"
    currency = "ethereum"

    # TODO: Migrate to __slots__
    METHODS = {
        "rpc_personal_new_account": "personal_newAccount",
        "rpc_eth_estimate_gas": "eth_estimateGas",
        "rpc_eth_gas_price": "eth_gasPrice",
        "rpc_personal_send_transaction": "personal_sendTransaction",
        "rpc_personal_unlock_account": "personal_unlockAccount",
        "rpc_eth_get_block_by_number": "eth_getBlockByNumber",
        "rpc_personal_list_accounts": "personal_listAccounts",
        "rpc_eth_get_transaction_by_hash": "eth_getTransactionByHash",
    }

    def __init__(
        self,
        rpc_host: str = "localhost",
        rpc_port: int = 18332,
        rpc_username: str = None,  # pylint: disable=unused-argument
        rpc_password: str = None,  # pylint: disable=unused-argument
        loop=None,
        session: aiohttp.ClientSession = None,
        timeout: int = None,
    ):
        self.auth = None
        self.headers = {
            "content-type": "application/json",
        }
        super().__init__(rpc_host, rpc_port, loop, session, timeout)

    async def wrapper(self, *args, method: str = None) -> Union[dict, list]:
        assert method is not None
        response = await self.call(
            payload={
                "method": method,
                "params": args,
                "jsonrpc": "2.0",
                "id": 1,
            }
        )
        return await self.validate(response)

    # Geth specific interface

    @staticmethod
    def calc_ether_fee(gas, gas_price):
        return utils.from_wei(utils.to_int(gas) * utils.to_int(gas_price))

    def format_transaction(self, tx, addresses):
        if tx["from"] in addresses and tx["to"] in addresses:
            category = "oneself"
        elif tx["from"] in addresses:
            category = "send"
        elif tx["to"] in addresses:
            category = "receive"
        else:
            assert False, "Unrecognized category"

        if tx["blockNumber"] is None:
            block_number = None
        else:
            block_number = utils.to_int(tx["blockNumber"])

        return {
            "txid": tx["hash"],
            "from_address": tx["from"],
            "to_address": tx["to"],
            "amount": utils.from_wei(utils.to_int(tx["value"])),
            "fee": self.calc_ether_fee(tx["gas"], tx["gasPrice"]),
            "block_number": block_number,
            "category": category,
            "timestamp": None,
            "info": tx,
        }

    async def fetch_blocks_range(
        self,
        start: int,
        end: int = None,
        bunch_size: int = 1000,
        delay: Union[int, float] = 0.1,
    ) -> List[dict]:
        """Fetches blocks range between start and end bounds.

        Args:
            start: Start fetching bound.
            end: End fetching bound (not inclusive). Defaults to
                latest block number.
            bunch_size: Concurrent RPC request number. Defaults to 1000.
            delay: Delay in seconds between concurrent request bunches.
                Defaults to 1.

        Returns:
            List that contains block range.
        """

        def to_bunches(coros, bunch_size):
            bunch_start = 0
            while bunch_start < len(coros):
                yield coros[bunch_start : bunch_start + bunch_size]
                bunch_start += bunch_size

        # TODO: Add validation

        end = end or await self.latest_block_number + 1
        get_block_coros = [
            self.rpc_eth_get_block_by_number(utils.to_hex(n), True)
            for n in range(start, end)
        ]
        blocks_range = []
        for bunch in to_bunches(get_block_coros, bunch_size):
            blocks_range += await asyncio.gather(*bunch)
            await asyncio.sleep(delay)
        return blocks_range

    async def fetch_recent_blocks_range(
        self,
        length: int,
        bunch_size: int = 1000,
        delay: Union[int, float] = 0.1,
    ):
        latest = await self.latest_block_number
        return await self.fetch_blocks_range(
            latest - length, latest + 1, bunch_size, delay
        )

    # Unified interface

    @property
    async def latest_block_number(self) -> int:
        latest_block = await self.rpc_eth_get_block_by_number("latest", True)
        return utils.to_int(latest_block["number"])

    async def create_address(self, password: str = "") -> str:
        return await self.rpc_personal_new_account(password)

    async def estimate_fee(
        self,  # pylint: disable=unused-argument
        transaction: dict = None,
        conf_target: int = 1,
    ) -> Decimal:
        if not transaction:
            raise TypeError(
                "Missing value for required keyword argument transaction"
            )
        if not isinstance(transaction, dict):
            raise TypeError(
                f"Transaction keyword argument must be a dict, "
                f"not '{type(transaction)}'"
            )
        gas_price = await self.rpc_eth_gas_price()
        estimated_gas = await self.rpc_eth_estimate_gas(transaction)
        return self.calc_ether_fee(estimated_gas, gas_price)

    async def send_transaction(
        self,
        amount: Decimal,
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
    ) -> dict:
        tx_data = {
            "from": from_address,
            "to": to_address,
            "value": utils.to_hex(utils.to_wei(amount)),
        }
        # TODO: Validate
        if isinstance(fee, dict):
            tx_data["gas"], tx_data["gasPrice"] = fee["gas"], fee["gas_price"]
        # TODO Execute in bunch
        addresses = await self.rpc_personal_list_accounts()
        txid = await self.rpc_personal_send_transaction(tx_data, password)
        tx = await self.rpc_eth_get_transaction_by_hash(txid)
        return self.format_transaction(tx, addresses)

    async def list_transactions(self, count: int = 10, **kwargs) -> List[dict]:
        """Lists most recent transactions.

        Args:
            count: The number of transactions to return. Defaults to 10.

        Returns:
            Recent transactions list.
        """

        def find_transactions_in(block, addresses):
            return [
                tx
                for tx in block["transactions"]
                if tx["from"] in addresses or tx["to"] in addresses
            ]

        bunch_size = kwargs.get("bunch_size", 1000)
        latest_block_number = await self.latest_block_number
        addresses = await self.rpc_personal_list_accounts()
        start = latest_block_number - bunch_size - 1
        end = latest_block_number + 1
        txs = []
        while True:
            blocks_range = await self.fetch_blocks_range(start, end, bunch_size)
            for block in blocks_range[::-1]:
                txs += find_transactions_in(block, addresses)
                if len(txs) >= count:
                    return [
                        self.format_transaction(tx, addresses)
                        for tx in txs[:count]
                    ]
            if start == 0:
                return [self.format_transaction(tx, addresses) for tx in txs]
            start -= bunch_size
            end -= bunch_size
            if start < 0:
                start = 0



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import asyncio
import functools
from decimal import Decimal
from typing import List, Union

import aiohttp

from obm import exceptions


def _catch_network_errors(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except aiohttp.ServerTimeoutError:
            self = args[0]
            raise exceptions.NetworkTimeoutError(
                f"The request to node was longer "
                f"than timeout: {self.timeout}"
            )
        except aiohttp.ClientError as exc:
            raise exceptions.NetworkError(exc)

    return wrapper


class Connector(abc.ABC):

    DEFAULT_TIMEOUT = 3

    def __init__(
        self, rpc_host, rpc_port, loop, session, timeout=DEFAULT_TIMEOUT
    ):
        # TODO: validate url
        if timeout is not None:
            if not isinstance(timeout, float):
                raise TypeError("Timeout must be a number")
            if timeout <= 0:
                raise ValueError("Timeout must be greater than zero")

        url = f"{rpc_host}:{rpc_port}"
        self.url = url if url.startswith("http") else "http://" + url
        self.timeout = aiohttp.ClientTimeout(total=timeout)
        self.loop = loop or asyncio.get_event_loop()
        self.session = session

    def __getattribute__(self, item):
        if item != "METHODS" and item in self.METHODS:
            return functools.partial(self.wrapper, method=self.METHODS[item])
        return super().__getattribute__(item)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def open(self):
        if self.session is None:
            self.session = aiohttp.ClientSession(
                loop=self.loop, headers=self.headers, auth=self.auth,
            )

    async def close(self):
        if self.session is not None:
            await self.session.close()
            self.session = None

    @_catch_network_errors
    async def call(self, payload: dict) -> dict:
        await self.open()
        async with self.session.post(
            url=self.url, json=payload, timeout=self.timeout
        ) as response:
            return await response.json()

    @staticmethod
    async def validate(response: dict) -> Union[dict, list]:
        try:
            if error := response.get("error"):
                raise exceptions.NodeError(error)
            return response["result"]
        except KeyError:
            raise exceptions.NodeInvalidResponceError(response)

    @property
    @abc.abstractmethod
    def node(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def currency(self) -> str:
        ...

    @abc.abstractmethod
    async def wrapper(self, *args, method: str = None) -> Union[dict, list]:
        ...

    # Unified interface

    @property
    @abc.abstractmethod
    async def latest_block_number(self) -> int:
        ...

    @abc.abstractmethod
    async def create_address(self, password: str = "") -> str:
        return await self.rpc_personal_new_account(password)

    @abc.abstractmethod
    async def estimate_fee(
        self, transaction: dict = None, conf_target: int = 1,
    ) -> Decimal:
        ...

    @abc.abstractmethod
    async def send_transaction(
        self,
        amount: Decimal,
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
    ) -> dict:
        ...

    @abc.abstractmethod
    async def list_transactions(self, count: int = 10, **kwargs) -> List[dict]:
        ...



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import marshmallow
from marshmallow import fields


class Transaction(marshmallow.Schema):
    txid = fields.String(required=True)
    from_address = fields.String(allow_none=True, required=True)
    to_address = fields.String(allow_none=True, required=True)
    amount = fields.Decimal(required=True)
    category = fields.String(allow_none=True, required=True)
    block_number = fields.Integer(allow_none=True, required=True)
    timestamp = fields.Integer(allow_none=True, required=True)
    fee = fields.Decimal(allow_none=True, required=True)
    info = fields.Dict(required=True)



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This magical module will rewrite all public methods in the public interface
of the library so they can run the loop on their own if it's not already
running. This rewrite may not be desirable if the end user always uses the
methods they way they should be ran, but it's incredibly useful for quick
scripts and the runtime overhead is relatively low.
"""
import asyncio
import functools
import inspect

from obm import mixins, models

__all__ = ["models", "mixins"]


def _syncify_wrap(_type, method_name):
    method = getattr(_type, method_name)

    @functools.wraps(method)
    def syncified(*args, **kwargs):
        coro = method(*args, **kwargs)
        loop = asyncio.get_event_loop()
        if loop.is_running():
            return coro
        return loop.run_until_complete(coro)

    # Save an accessible reference to the original method
    syncified.asynchronous = method
    setattr(_type, method_name, syncified)


def syncify(*types):
    """Converts all async methods in given types into synchronous.

    Converted methods return either the coroutine or the result
    based on whether asyncio's event loop is running.
    """
    for _type in types:
        for name in dir(_type):
            if not name.startswith("_") or name == "__call__":
                if inspect.iscoroutinefunction(getattr(_type, name)):
                    _syncify_wrap(_type, name)


syncify(mixins.ConnectorMixin)



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools

import obm


def read(file_name):
    with open(file_name) as f:
        content = f.read()
    return content


setuptools.setup(
    name="obm",
    version=obm.__version__,
    packages=setuptools.find_packages(exclude=["tests*"]),
    install_requires=["aiohttp>=3.6,<4", "web3>=5.7,<6", "marshmallow>=3.5,<4"],
    extras_require={
        "dev": [
            "sphinx>=2.4,<3",
            "sphinx-rtd-theme",
            "python-dotenv",
            "pytest",
            "pylint",
            "mypy",
            "rope",
        ],
    },
    license="GNU Lesser General Public License v3 or later (LGPLv3+)",
    description="Async blockchain nodes interacting tool with ORM-like api.",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    url="https://github.com/madnesspie/obm",
    author="Alexander Polishchuk",
    author_email="apolishchuk52@gmail.com",
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
)
